import gevent.monkey
gevent.monkey.patch_all()

import socket, gevent, json, node, uidlib, gevent.server
from protocol import Protocol, Connect, LoopBackProtocol
from finger import FingerTable
from set_store import SetHandler

class NetworkListener():
    def __init__(self, start_addr, port=8339, ip='127.0.0.1'):
        #Special node which talks about ourselves
        self.port = port
        self.ip = ip
        self.node = node.Node(uidlib.new_uid(), ip, port, None)
        
        #Handler Classes
        self.finger = FingerTable(self.node)
        self.set_handler = SetHandler(self.finger)
        
        #Set up the listening server
        self._s= gevent.server.StreamServer(( self.ip, self.port), self._handle)
        self._s.start()
        
        gevent.spawn(self._ask_help)
        
        if start_addr:
            Connect(self.finger, self.set_handler, start_addr)
        LoopBackProtocol(ip + ":" + str(port), self.finger, self.set_handler)
                
    def _handle(self, conn, addr):
        Protocol(conn, ":".join((addr[0], str(addr[1]))), self.finger, self.set_handler)
        
    def _ask_help(self):
        while True:
            needed = self.finger.get_levels()
            for level in needed:
                self.finger.level_send(level, 'REQ_LEVEL ' + str(level))
            gevent.sleep(60)
        
import unittest

class TestNetwork(unittest.TestCase):
    def testConnection(self):
        net = NetworkListener(None)
        net2 = NetworkListener(net.node.addr, port = 8337)
        with gevent.Timeout(1) as timeout:
            while len(net.finger.known) <2:
                gevent.sleep()
        self.assertTrue(len(net.finger.known) == 2)
        gevent.sleep(0.05)
        self.assertTrue(len(net.finger.known) == 2)
        
if __name__ == "__main__":
    unittest.main()  
                



"""
Handles uid creation and conversion to integer types
"""

from hashlib import md5
from os import urandom

def new_uid():
    a = md5()
    a.update(urandom(16))
    return a.hexdigest()
    
def uid_2_num(uid):
    return long(uid, 16)
    
def num_2_uid(num):
    item = hex(num)[2:-1]
    return (32 - len(item)) * '0' + item
    
def distance(a, b):
    if isinstance(a, str):
        a = uid_2_num(a)
    if isinstance(b, str):
        b = uid_2_num(b)
    return a^b
    
import unittest

class uidtest(unittest.TestCase):
    def testTransforms(self):
        for i in range(100):
            a = new_uid()
            self.assertEqual(a, num_2_uid(uid_2_num(a)))
            
if __name__ == "__main__":
    unittest.main()




"""
Class to store node information.
The actual class which sends stuff is the protocol class
"""

import uidlib, time

class Node():
    """ Contains contact info """
    def __init__(self, uid, host, port, prot):
        self.uid = uid
        self.num = uidlib.uid_2_num(uid)
        self.host = host
        self.port = port
        self.prot = prot
        self.last_seen = -1
        self.addr = host +":" + str(port)
        
    def send(self, msg):
        self.prot.send(msg)
        
    def seen(self):
        self.last_seen = time.time()
        
    def __repr__(self):
        return '<Node ' + str((self.uid, self.host, self.port, self.prot)) + ' >'
        
    def __lt__(self, a):
        if isinstance(a, Node):
            a = a.num
        return self.num < a
        
    def __le__(self, a):
        if isinstance(a, Node):
            a = a.num
        return self.num <= a
        
    def __gt__(self, a):
        if isinstance(a, Node):
            a = a.num
        return self.num > a
    
    def __ge__(self, a):
        if isinstance(a, Node):
            a = a.num
        return self.num >= a
        
    def __eq__(self, a):
        if isinstance(a, Node):
            a = a.num
        return self.num == a
        
    def __ne__(self, a):
        if isinstance(a, Node):
            a = a.num
        return self.num != a
        
    def __hash__(self):
        return hash(self.uid) ^ hash(self.host) ^ hash(self.port)
        
        

import unittest
class TestNode(unittest.TestCase):
    def testCreation(self):
        a = Node(uidlib.new_uid(), 'localhost', '8338', None)
        self.assertTrue(isinstance(a, Node))
        
    def testHashabilitySet(self):
        import copy
        a = Node(uidlib.new_uid(), 'localhost', '8338', None)
        b = set([])
        b.add(a)
        c = copy.copy(a)
        b.add(c)
        self.assertTrue(len(b) == 1)
        
if __name__ == "__main__":
    unittest.main()



import uidlib, math, gevent

class FingerTable():
    def __init__(self, self_node, min_count=1):
        self.known = set([])
        self.self = self_node
        self.table = []
        self.min_count = min_count
        self.max_level = int(self._uid_2_level(32*'F', 32*'0'))+1
        for i in range(self.max_level):
            self.table.append(set([]))
        
    def _above_zero(self, item):
        if item < 0: 
            return 0
        return item
        
    def _uid_2_level(self, uid, other_uid=None):
        """
        Convert a uid to a log level in the finger table
        """
        if other_uid == None:
            other_uid = self.self.uid
        level = int(self._above_zero(
                        math.log(
                            uidlib.distance(uid, other_uid)+0.1)
                        )
                    )
        return level
        
    def _level_check(self, level):
        """
        Make sure the level is inside the table range.
        """
        if level > len(self.table) - 1:
            for i in range(len(self.table)-1, level):
                self.table.append(set([]))
                self.max_level += 1
                
    def add(self, Node):
        """
        Add a node to our finger table
        """
        self.known.add(Node)
        level = self._uid_2_level(Node.uid)
        self._level_check(level)
        self.table[level].add(Node)
        
    def remove(self, Node):
        """
        remove a node from our finger table
        """
        if Node not in self.known:
            return
        self.known.remove(Node)
        level = self._uid_2_level(Node.uid)
        self.table[level].remove(Node)
        
    def get_levels(self):
        """
        return what distances we would like to get uids for
        """
        for i in range(len(self.table)):
            if len(self.table[i]) < self.min_count:
                yield int(i)
                
    def get_node_from_level(self, level, uid):
        """
        Gets the node which is a certain level away from another uid
        """
        for i in xrange(
            int(level-self._uid_2_level(uid)), 
            int(level+self._uid_2_level(uid))
            ): #Triangle Inequality
            if i < len(self.table) and i >=0:
                for node in self.table[i]:
                    if self._uid_2_level(node.uid, uid) == level:
                        return node
        return None
            
                
    def level_send(self, level, msg):
        """
        Send message to the node closest to that level
        """
        self._level_check(level)
        closest_node = None
        closest_level = 0
        for node in self.known:
            node_level = self._uid_2_level(node.uid)
            if not closest_node or abs(node_level - level) < closest_level:
                closest_node = node
                closest_level = abs(node_level - level)
            #Do a cooperative yield since this could take a while
            #gevent.sleep()
        if closest_node:
            closest_node.prot.send(msg)
                          
    def get_node(self, uid):
        """
        get closest node
        """
        level = self._uid_2_level(uid)
        self._level_check(level)
        
        #Check if we have an entry
        if len(self.table[level]):
            search = self.table[level]
        #Else find the closest node
        else:
            search = self.known
            
        close_node = None
        for node in search:
            if not close_node or uidlib.distance(node.uid, uid) < uidlib.distance(close_node.uid, uid):
                close_node = node
            
        return close_node  
               
    def send(self, uid, msg):
        """
        Sends to the node closest to that uid
        """
        
        node = self.get_node(uid)
        if node and node.prot:  
            node.prot.send(msg)
        
import unittest, node
Node = node.Node
class fingertest(unittest.TestCase):
    def testadds(self):
    
        finger = FingerTable(Node(uidlib.new_uid(), '', '', ''))
        Nodes = set([])
        
        for i in range(100):
            a = uidlib.new_uid()
            b = Node(a, '', '', '')
            Nodes.add(b)
            finger.add(b)
            
        for y in finger.get_levels():
            pass
            
        for x in Nodes:
            finger.remove(x)
            
        self.assertTrue(len(finger.known) == 0)
        
    def test_levels_needed(self):
        finger = FingerTable(Node(uidlib.new_uid(), '', '', ''))
        self.assertTrue(len([x for x in finger.get_levels()]) == finger.max_level)
        finger.add(Node(uidlib.new_uid(), '', '', ''))
        self.assertTrue(len([x for x in finger.get_levels()]) == finger.max_level-1)
            
            
if __name__ == "__main__":
    unittest.main()





import gevent.monkey
gevent.monkey.patch_all()

import json, gevent.queue, gevent

class SetHandler():
    def __init__(self, finger):
        self.finger = finger
        self.dict = {}
        self.queue = {}
        gevent.spawn(self.check_updates)
        
    def check_updates(self):
        known = len(self.finger.known)
        while True:
            if len(self.finger.known) != known:
                known = len(self.finger.known)
                for k in self.dict:
                    self.update(k,self.dict[k])
            gevent.sleep(0.1)
        
    def get(self, key):
        for x in range(3):
            try:
                self.queue[key] = gevent.queue.Queue(0)
                self.finger.send(hash(key) % int(32 * 'F', 16), 'GET ' + json.dumps(key))
                with gevent.Timeout(0.5):
                    result = self.queue[key].get()
                    break
            except gevent.Timeout:
                result = None
                continue
        del self.queue[key]
        return result
        
    def set(self, key, value):
        self.finger.send(hash(key) % int(32 * 'F', 16), 'SET ' + json.dumps([key, value]))
        
    def update(self, key, value):
        self.finger.send(hash(key) % int(32 * 'F', 16), 'UPD ' + json.dumps([key, value]))
        
    def add(self, key, value):
        self.finger.send(hash(key) % int(32 * 'F', 16), 'ADD ' + json.dumps([key, value]))
        
    def handle_msg(self, proto, msg):
        if msg[0:3] not in ['GET', 'SET', 'ADD', 'RES', 'UPD']:
            return
            
        if msg[0:3] in ['GET']:
            _, key = msg.split(' ', 1)
            key = json.loads(key)
            if key in self.dict:
                item = self.dict[key]
                proto.send('RES ' +  json.dumps([key,item]))
            else:
                proto.send('RES ' + json.dumps([key, None]))
            
        if msg[0:3] in ['SET']:
            _, item = msg.split(' ', 1)
            key, value = json.loads(item)
            self.dict[key] = value
            
        if msg[0:3] in ['ADD']:
            _, item = msg.split(' ', 1)
            key, value = json.loads(item)
            if key not in self.dict:
                self.dict[key] = []
            self.dict[key].append(value)
            
        if msg[0:3] in ['RES']:
            _, item = msg.split(' ', 1)
            key, value = json.loads(item)
            if key in self.queue:
                self.queue[key].put(value)
            
        if msg[0:3] in ['UPD']:
            _, item = msg.split(' ', 1)
            key, value = json.loads(item)
            if key not in self.dict or self.dict[key] is not value:
                self.dict[key] = value
    
import unittest

class TestStore(unittest.TestCase):
    def testConnection(self):
        gevent.monkey.patch_all()
        from network import NetworkListener

        net = NetworkListener(None, port = 8345)
        netlist = [net]
        for i in range(100):
            new = NetworkListener(net.node.addr, port = 8346+i)
            netlist.append(new)

        while len(net.finger.known) < 2:
            gevent.sleep()

        net.set_handler.set('hi','bar')
        net.set_handler.set('lp_blah', ['hi'])
        net.set_handler.add('lp_blah', 'bar')

        with gevent.Timeout(3):
            while sum(map(lambda x: len(x.set_handler.dict), netlist)) < 2:
                gevent.sleep()
        self.assertEqual(net.set_handler.get('hi'), 'bar')
        self.assertEqual(net.set_handler.get('lp_blah'), ['hi','bar'])

if __name__ == "__main__":
    unittest.main()  



#!/usr/bin/env python

from setuptools import setup
setup(name='gevent_dht',
        version='0.1.1',
        description='Gevent based distributed hash table',
        author='Colin Rice',
        author_email='dah4k0r@gmail.com',
        packages = ['gevent_dht'],
        url='github.com/c00w/gevent_dht',
        classifiers=[
            'Development Status :: 4 - Beta',
            'Intended Audience :: Developers',
            'Topic :: System :: Networking',
            'License :: OSI Approved :: MIT License',
            ],
        install_requires= 'gevent',
        license = 'MIT Expat License'
        )



import gevent
import network
import socket
import uidlib

class distributedHashTable():
    def __init__(self, first_node, local_port=8339, local_ip=''):
        """
        First_node is the address of the first computer to connect to
        If it is none then you are bootstrapping the network and this
        is the first node.
        
        local_port is the local port to bind to.
        Default 8339.
        
        local_ip is the interface to bind to.
        Default is '' which should bind all addrresses
        """
        
        self.listener = network.NetworkListener(first_node, local_port, local_ip)
        
    def __get_node_from_key(self, key):
        #Get its hash mod the id limit
        key_hash = hash(key) % int('F'*32, 16)
        node = self.listener.finger.get_node(key_hash)
        
    def __getitem__(self, key):
        """
        Get an item from the distributed hash table
        """
        
        node = self.__get_node_from_key(key)
        return node.set_handler.get(key)
        
    def __setitem__(self, key, value):
        """
        Set an item in the distributed hash table
        """
        
        node = self.__get_node_from_key(key)
        return node.set_handler.set(key, value)
        
    def append(self, key, value):
        """
        Remove item
        """
        node = self.__get_node_from_key(key)
        return node.set_handler.add(key, value)
        



import gevent, gevent.queue, socket, time
from node import Node

def addr_2_host_port(addr):
    host = ":".join(addr.split(':')[:-1])
    port = addr.split(':')[-1]
    return host, int(port)

def Connect(finger, set_handler, addr):
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.connect((addr_2_host_port(addr)))
    return Protocol(s, addr, finger, set_handler)

class Protocol():
    def __init__(self, conn, addr, finger, set_handler):
        self.remote_addr = addr
        self.remote_conn = conn
        self.finger = finger
        self.Node = None
        self.set_handler = set_handler
        
        #Queues and generators for message handling
        self.recv_gen = self.recv_generator(conn)
        self.send_queue = gevent.queue.Queue()
        
        #Get address and port
        self.host, self.port = addr_2_host_port(addr)
        
        #Spawn handler threads.
        gevent.spawn(self.local_handle)
        gevent.spawn(self.net_handle)
        gevent.spawn(self.check_alive)
        
        self.send('UIDRESP ' + self.finger.self.uid)
        
    def check_alive(self):
        while True:
            gevent.sleep(60*1)
            if (time.time() - self.Node.last_seen) > 60:
                self.send('PING')
            if (time.time() - self.Node.last_seen) > 60*5:
                if self.Node:
                    self.finger.remove(self.Node)
                del self
                return
        
    def send(self, msg):
        self.send_queue.put(msg)
        
    def recv_generator(self, conn):
        msg = ''
        while True:
            new_data = conn.recv(32)
            msg += new_data
            if '|' in msg:
                length, remainder = msg.split('|', 1)
                length = int(length)
                if len(remainder) < length:
                    continue
                
                yield remainder[0:length]
                msg = remainder[length:len(remainder)]
        
    def net_handle(self):
        "Handle remote network messages"
        for msg in self.recv_gen:
            self.net_msg_handle(msg)
            
    def local_handle(self):
        "Handle requests for messages to be sent"
        for item in self.send_queue:
            self.net_msg_send(item)
            
    def net_msg_send(self, item):
        "Clean messages so they don't mess up the protocol"
        #print 'SENT: ' + item
        msg = str(len(item)) + '|' + item
        self.remote_conn.send(msg)
        
    def net_msg_handle(self, msg):
        if msg[0:7] in ['UIDRESP']:
            _, uid = msg.split(' ', 1)
            if self.Node == None:
                self.Node = Node(uid, self.host, self.port, self)
            self.finger.add(self.Node)
            
        if msg[0:6] in ['UIDREQ']:
            resp = 'UIDRESP ' + self.finger.self.uid
            self.send(resp)
            
        if msg[0:4] in ['PING']:
            self.send('PONG')
        
        if msg[0:4] in ['PONG']:
            pass
            
        if self.Node == None:
            self.send('UIDREQ')
            return
        else:
            self.Node.seen()
            
        if msg[0:9] in ['REQ_LEVEL']:
            _, level = msg.split(' ', 1)
            node = self.finger.get_node_from_level(int(level), self.Node.uid)
            if node:
                resp = 'RESP_LEVEL ' + str(node.uid) + " " + self.host + ":" + str(self.port)
                self.send(resp)
                
        if msg[0:10] in ['RESP_LEVEL']:
            _, uid, addr = msg.split(' ', 1)
            Connect(self.finger, self.set_handler, addr)
            
        self.set_handler.handle_msg(self, msg)
            
    def __del__(self):
        self.send_queue.put(StopIteration)
        self.remote_conn.shutdown()
        self.remote_conn.close()
        
        
class LoopBackProtocol(Protocol):
    """
    Protocol for talking to yourself
    """
    def __init__(self, addr, finger, set_handler):
        self.remote_addr = addr
        self.finger = finger
        self.Node = None
        self.set_handler = set_handler
        
        #Queues and generators for message handling
        self.recv_gen = gevent.queue.Queue()
        self.send_queue = gevent.queue.Queue()
        
        #Get address and port
        self.host, self.port = addr_2_host_port(addr)
        
        #Spawn handler threads.
        gevent.spawn(self.local_handle)
        gevent.spawn(self.net_handle)
        gevent.spawn(self.check_alive)
        
        self.send('UIDRESP ' + self.finger.self.uid)
        
    def net_msg_send(self, item):
        "Send messages locally so we can talk to ourselves"
        self.recv_gen.put(item)
        
import unittest

class TestProtocol(unittest.TestCase):
    def testaddr(self):
        self.assertTrue(addr_2_host_port("127.0.0.1:80") == ("127.0.0.1", 80))
        self.assertTrue(addr_2_host_port("fe80::1234:80") == ("fe80::1234", 80))
       
if __name__ == "__main__":
    unittest.main()
